# Copyright 2020 Lorna Authors. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# Refrence https://github.com/sbarratt/inception-score-pytorch/blob/master/inception_score.py

import numpy as np
import torch
import torch.nn as nn
import torch.utils.data
import torchvision.datasets as dset
import torchvision.transforms as transforms
from scipy.stats import entropy
from torch.autograd import Variable
from torch.nn import functional as F


class LeNet(nn.Module):
    def __init__(self):
        super(LeNet, self).__init__()
        self.conv1 = nn.Sequential(
            nn.Conv2d(1, 6, 5, 1, 2),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=2, stride=2),
        )
        self.conv2 = nn.Sequential(
            nn.Conv2d(6, 16, 5),
            nn.ReLU(),
            nn.MaxPool2d(2, 2)
        )
        self.fc1 = nn.Sequential(
            nn.Linear(16 * 5 * 5, 120),
            nn.ReLU()
        )
        self.fc2 = nn.Sequential(
            nn.Linear(120, 84),
            nn.ReLU()
        )
        self.fc3 = nn.Linear(84, 10)

    def forward(self, x):
        x = self.conv1(x)
        x = self.conv2(x)

        x = torch.flatten(x, 1)

        x = self.fc1(x)
        x = self.fc2(x)
        x = self.fc3(x)
        return x


class IgnoreLabelDataset(torch.utils.data.Dataset):
    def __init__(self, orig):
        self.orig = orig

    def __getitem__(self, index):
        return self.orig[index][0]

    def __len__(self):
        return len(self.orig)


def inception_score(images, batch_size=32, splits=1):
    """ Computes the inception score of the generated images

    Args:
        images (tensor): Torch dataset of (3xHxW) numpy images normalized in the range [-1, 1]
        batch_size (int): Batch size for feeding into LeNet.
        splits (int): Number of splits.

    Returns:
        Inception Score.
    """
    images_num = len(images)

    assert batch_size > 0
    assert images_num > batch_size

    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    # Set up dataloader
    dataloader = torch.utils.data.DataLoader(images, batch_size=batch_size)

    # Load inception model
    model = LeNet().to(device)
    state = torch.load("../research/mnist.pth")
    model.load_state_dict(state)
    model.eval()

    def get_pred(x):
        x = model(x)
        return F.softmax(x).data.cpu().numpy()

    # Get predictions
    preds = np.zeros((images_num, 10))

    for i, batch in enumerate(dataloader, 0):
        batch = batch
        batch = batch.to(device)
        batchv = Variable(batch)
        batch_size_i = batch.size()[0]

        preds[i * batch_size:i * batch_size + batch_size_i] = get_pred(batchv)

    # Now compute the mean kl-div
    split_scores = []

    for k in range(splits):
        part = preds[k * (images_num // splits): (k + 1) * (images_num // splits), :]
        py = np.mean(part, axis=0)
        scores = []
        for i in range(part.shape[0]):
            pyx = part[i, :]
            scores.append(entropy(pyx, py))
        split_scores.append(np.exp(np.mean(scores)))

    return np.mean(split_scores), np.std(split_scores)


if __name__ == '__main__':
    dataset = dset.MNIST(root='../research/data/',
                         download=True,
                         transform=transforms.Compose([
                             transforms.ToTensor(),
                             transforms.Normalize((0.5,), (0.5,))
                         ]))

    print("Calculating Inception Score...")
    print(inception_score(IgnoreLabelDataset(dataset), batch_size=32, splits=10))



# Copyright 2020 Lorna Authors. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import collections
import hashlib
import os
import ssl

import torch
import torch.utils.model_zoo as model_zoo

# Parameters for the entire model (stem, all blocks, and head)
GlobalParams = collections.namedtuple("GlobalParams", [
    "noise", "channels", "image_size",
    "batch_norm_momentum", "negative_slope"
])

# Change namedtuple defaults
GlobalParams.__new__.__defaults__ = (None,) * len(GlobalParams._fields)


########################################################################
############## HELPERS FUNCTIONS FOR LOADING MODEL PARAMS ##############
########################################################################


def model_params(model_name):
    r""" Map Generator and Discriminator model name to parameter coefficients.

      Args:
        model_name (string): The name of the model corresponding to the dataset.

      Returns:
        A binary tuple value.
      """

    params_dict = {
        # Coefficients: channels, image_size
        "g-mnist": (1, 28),
        "g-fmnist": (1, 28),
        "d-mnist": (1, 28),
        "d-fmnist": (1, 28),
    }
    return params_dict[model_name]


def model(channels=None, image_size=None):
    r""" Gets the parameters of the model

      Args:
        channels (int): size of each input image channels.
        image_size (int): size of each input image size.

      Returns:
        A set of GlobalParams shared between blocks.
      """

    global_params = GlobalParams(
        noise=100,
        batch_norm_momentum=0.8,
        negative_slope=0.2,
        channels=channels,
        image_size=image_size,
    )

    return global_params


def get_model_params(model_name):
    """ Get the block args and global params for a given model

      Args:
        model_name (string): The name of the model corresponding to the dataset.

      Returns:
        A set of GlobalParams shared between blocks.
      """

    if model_name.startswith("g") or model_name.startswith("d"):
        c, s = model_params(model_name)
        global_params = model(channels=c, image_size=s)
    else:
        raise NotImplementedError(f"model name is not pre-defined: {model_name}.")
    return global_params


urls_map = {
    "d-fmnist": "https://github.com/Lornatang/GAN-PyTorch/releases/download/1.0/d-fmnist-68a00804.pth",
    "d-mnist": "https://github.com/Lornatang/GAN-PyTorch/releases/download/1.0/d-mnist-73aa2235.pth",
    "g-fmnist": "https://github.com/Lornatang/GAN-PyTorch/releases/download/1.0/g-fmnist-5036734e.pth",
    "g-mnist": "https://github.com/Lornatang/GAN-PyTorch/releases/download/1.0/g-mnist-a0922495.pth",
}


def load_pretrained_weights(model_arch, model_name):
    """ Loads pretrained weights, and downloads if loading for the first time. """

    state_dict = model_zoo.load_url(urls_map[model_name])
    model_arch.load_state_dict(state_dict)
    print(f"Loaded model pretrained weights for `{model_name}`.")



# Copyright 2020 Lorna Authors. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import hashlib
import os

import torch


def cal_file_md5(filename):
    """ Calculates the MD5 value of the file
    Args:
        filename: The path name of the file.

    Return:
        The MD5 value of the file.

    """
    with open(filename, "rb") as f:
        md5 = hashlib.md5()
        md5.update(f.read())
        hash_value = md5.hexdigest()
    return hash_value


def compress_model(state, filename, model_arch):
    model_folder = "../checkpoints"
    try:
        os.makedirs(model_folder)
    except OSError:
        pass

    new_filename = model_arch + "-" + cal_file_md5(filename)[:8] + ".pth"
    torch.save(state, os.path.join(model_folder, new_filename))



# Copyright 2020 Lorna Authors. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import torch
import torch.nn as nn

from .utils import get_model_params
from .utils import load_pretrained_weights
from .utils import model_params


# Generative Adversarial Networks model architecture from the One weird trick...
# <https://arxiv.org/abs/1406.2661>`_ paper.
class Generator(nn.Module):
  r""" An Generator model. Most easily loaded with the .from_name or .from_pretrained methods

  Args:
    global_params (namedtuple): A set of GlobalParams shared between blocks

  Examples:
      >>> import torch
      >>> from gan_pytorch import Generator
      >>> from gan_pytorch import Discriminator
      >>> generator = Generator.from_pretrained("g-mnist")
      >>> discriminator = Discriminator.from_pretrained("g-mnist")
      >>> generator.eval()
      >>> discriminator.eval()
      >>> noise = torch.randn(1, 100)
      >>> discriminator(generator(noise)).item()
  """

  def __init__(self, global_params=None):
    super(Generator, self).__init__()
    self.channels = global_params.channels
    self.image_size = global_params.image_size

    self.main = nn.Sequential(
      nn.Linear(global_params.noise, 128),
      nn.LeakyReLU(global_params.negative_slope, inplace=True),

      nn.Linear(128, 256),
      nn.BatchNorm1d(256, global_params.batch_norm_momentum),
      nn.LeakyReLU(global_params.negative_slope, inplace=True),

      nn.Linear(256, 512),
      nn.BatchNorm1d(512, global_params.batch_norm_momentum),
      nn.LeakyReLU(global_params.negative_slope, inplace=True),

      nn.Linear(512, 1024),
      nn.BatchNorm1d(1024, global_params.batch_norm_momentum),
      nn.LeakyReLU(global_params.negative_slope, inplace=True),

      nn.Linear(1024, self.channels * self.image_size * self.image_size),
      nn.Tanh()
    )

  def forward(self, x):
    r"""Defines the computation performed at every call.

    Args:
      x (tensor): input tensor into the calculation.

    Returns:
      A four-dimensional vector (NCHW).
    """
    x = self.main(x)
    x = x.reshape(x.size(0), self.channels, self.image_size, self.image_size)
    return x

  @classmethod
  def from_name(cls, model_name):
    cls._check_model_name_is_valid(model_name)
    global_params = get_model_params(model_name)
    return cls(global_params)

  @classmethod
  def from_pretrained(cls, model_name):
    model = cls.from_name(model_name, )
    load_pretrained_weights(model, model_name)
    return model

  @classmethod
  def get_image_size(cls, model_name):
    cls._check_model_name_is_valid(model_name)
    _, res = model_params(model_name)
    return res

  @classmethod
  def _check_model_name_is_valid(cls, model_name):
    """ Validates model name. """
    valid_list = ["mnist", "fmnist"]
    valid_models = ["g-" + str(i) for i in valid_list]
    if model_name not in valid_models:
      raise ValueError("model_name should be one of: " + ", ".join(valid_models))


class Discriminator(nn.Module):
  r""" An Discriminator model. Most easily loaded with the .from_name or .from_pretrained methods

  Args:
    global_params (namedtuple): A set of GlobalParams shared between blocks

  Examples:
    >>> import torch
    >>> from gan_pytorch import Discriminator
    >>> discriminator = Discriminator.from_pretrained("d-mnist")
    >>> discriminator.eval()
    >>> noise = torch.randn(1, 784)
    >>> discriminator(noise).item()
  """

  def __init__(self, global_params=None):
    super(Discriminator, self).__init__()

    self.main = nn.Sequential(
      nn.Linear(global_params.channels * global_params.image_size * global_params.image_size, 512),
      nn.LeakyReLU(global_params.negative_slope, inplace=True),

      nn.Linear(512, 256),
      nn.LeakyReLU(global_params.negative_slope, inplace=True),

      nn.Linear(256, 1),
      nn.Sigmoid(),
    )

  def forward(self, x):
    r""" Defines the computation performed at every call.

    Args:
      x (tensor): input tensor into the calculation.

    Returns:
      A four-dimensional vector (NCHW).
    """
    x = torch.flatten(x, 1)
    x = self.main(x)
    return x

  @classmethod
  def from_name(cls, model_name):
    cls._check_model_name_is_valid(model_name)
    global_params = get_model_params(model_name)
    return cls(global_params)

  @classmethod
  def from_pretrained(cls, model_name):
    model = cls.from_name(model_name)
    load_pretrained_weights(model, model_name)
    return model

  @classmethod
  def get_image_size(cls, model_name):
    cls._check_model_name_is_valid(model_name)
    _, res = model_params(model_name)
    return res

  @classmethod
  def _check_model_name_is_valid(cls, model_name):
    """ Validates model name. """
    valid_list = ["mnist", "fmnist"]
    valid_models = ["d-" + str(i) for i in valid_list]
    if model_name not in valid_models:
      raise ValueError("model_name should be one of: " + ", ".join(valid_models))
